﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis.Symbolic {
  [StorableClass]
  public sealed class SymbolicDataAnalysisIslandGAEvaluator<T> : SingleSuccessorOperator, IStochasticOperator, ISymbolicDataAnalysisIslandGAEvaluator
    where T : class,IDataAnalysisProblemData {
    private const string RandomParameterName = "Random";
    private const string ProblemDataParameterName = "ProblemData";
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string EvaluatorParameterName = "ProblemEvaluator";
    private const string QualityParameterName = "Quality";
    private const string FitnessCalculationPartitionParameterName = "FitnessCalculationPartition";
    private const string FixedSamplesPartitionParameterName = "FixedSamplesPartition";
    private const string RandomSamplesParameterName = "RandomSamples";

    #region parameter properties
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters[RandomParameterName]; }
    }
    public ILookupParameter<T> ProblemDataParameter {
      get { return (ILookupParameter<T>)Parameters[ProblemDataParameterName]; }
    }
    public ILookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ILookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    public ILookupParameter<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>> EvaluatorParameter {
      get { return (ILookupParameter<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>>)Parameters[EvaluatorParameterName]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters[QualityParameterName]; }
    }
    public IValueLookupParameter<IntRange> FitnessCalculationPartitionParameter {
      get { return (IValueLookupParameter<IntRange>)Parameters[FitnessCalculationPartitionParameterName]; }
    }
    public ILookupParameter<IntRange> FixedSamplesPartitionParameter {
      get { return (ILookupParameter<IntRange>)Parameters[FixedSamplesPartitionParameterName]; }
    }
    public ILookupParameter<IntValue> RandomSamplesParameter {
      get { return (ILookupParameter<IntValue>)Parameters[RandomSamplesParameterName]; }
    }

    public IValueParameter<EnumerableItem<int>> RowsParameter {
      get { return (IValueParameter<EnumerableItem<int>>)Parameters["Rows"]; }
    }
    #endregion


    [StorableConstructor]
    private SymbolicDataAnalysisIslandGAEvaluator(bool deserializing) : base(deserializing) { }
    private SymbolicDataAnalysisIslandGAEvaluator(SymbolicDataAnalysisIslandGAEvaluator<T> original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicDataAnalysisIslandGAEvaluator<T>(this, cloner);
    }

    public SymbolicDataAnalysisIslandGAEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>(RandomParameterName, "The random generator to use."));
      Parameters.Add(new LookupParameter<T>(ProblemDataParameterName, "The problem data on which the symbolic data analysis solution should be evaluated."));
      Parameters.Add(new LookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic data analysis solution encoded as a symbolic expression tree."));
      Parameters.Add(new LookupParameter<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>>(EvaluatorParameterName, "The evaluator provided by the symbolic data analysis  problem."));
      Parameters.Add(new LookupParameter<DoubleValue>(QualityParameterName, "The quality which is calculated by the encapsulated evaluator."));
      Parameters.Add(new ValueLookupParameter<IntRange>(FitnessCalculationPartitionParameterName, "The data partition used to calculate the fitness"));
      Parameters.Add(new LookupParameter<IntRange>(FixedSamplesPartitionParameterName, "The data partition which is used to calculate the fitness on the fixed samples."));
      Parameters.Add(new LookupParameter<IntValue>(RandomSamplesParameterName, "The number of random samples used for fitness calculation in each island."));
      Parameters.Add(new OptionalValueParameter<EnumerableItem<int>>("Rows", "TODO"));

      EvaluatorParameter.Hidden = true;
    }

    public override IOperation Apply() {
      var evaluator = EvaluatorParameter.ActualValue;
      var problemData = ProblemDataParameter.ActualValue;

      var samples = FitnessCalculationPartitionParameter.ActualValue;
      var fixedSamples = FixedSamplesPartitionParameter.ActualValue;
      var randomSamples = RandomSamplesParameter.ActualValue.Value;

      //create fixed rows enumerable
      var rows = Enumerable.Range(fixedSamples.Start, fixedSamples.Size);
      //create randomly chosen rows enumerable
      if (randomSamples > 0) {
        if (randomSamples > samples.Size - fixedSamples.Size) {
          var error = string.Format("Could not select {0} random samples, because there are {1} total samples present from which {2} where used in the fixed partition. Please lower the number of random samples in the algorithm configuration.", randomSamples, samples.Size, fixedSamples.Size);
          throw new OperatorExecutionException(this, error);
        }
        var randomRows = Enumerable.Range(samples.Start, samples.Size).Where(r => r < fixedSamples.Start || r >= fixedSamples.End);
        randomRows = randomRows.SampleRandomWithoutRepetition(RandomParameter.ActualValue, randomSamples, samples.Size - fixedSamples.Size);

        rows = rows.Concat(randomRows);
      }
      //filter out test rows
      rows = rows.Where(r => r < problemData.TestPartition.Start && r > problemData.TestPartition.End);

      //execution context is created manually to be able to clear the rows parameter easily
      RowsParameter.Value = new EnumerableItem<int>(rows);
      var executionContext = new ExecutionContext(ExecutionContext, evaluator, ExecutionContext.Scope);
      var successor = evaluator.Execute(executionContext, this.CancellationToken);
      RowsParameter.Value = null;
      return new OperationCollection(successor, base.Apply());
    }
  }
}
